@@ -1,17 +1,12 @@
+/*   This has nothing to do with cryptography. */
 #include <config.h>
-#include <errno.h>
-#include <sys/types.h>
-#include <malloc.h>
-#include <unistd.h>
-#include <stdio.h>
-#include <string.h>
-#include "mad.h"
-#include "fs.h"
-
-#include <fcntl.h>
+#include "net-includes.h"
+#include "diffie/diffie-socket.h"
+#include "diffie/z-socket.h"
 #include "util.h"
-
 #include "vfs.h"
+#include "diffie/compat.h"
+#include "mad.h"
 
     
 static void *local_open (char *file, int flags, int mode)
@@ -69,7 +64,6 @@ static void *local_opendir (char *dirname)
 {
     DIR **local_info;
     DIR *dir;
-
     dir = opendir (dirname);
     if (!dir)
 	return 0;
@@ -82,7 +76,9 @@ static void *local_opendir (char *dirname)
 
 static void *local_readdir (void *data)
 {
-    return readdir (*(DIR **) data);
+    struct dirent *r;
+    r = readdir (*(DIR **) data);
+    return r;
 }
 
 static int local_closedir (void *data)
@@ -114,7 +110,7 @@ static int local_fstat (void *data, struct stat *buf)
     return fstat (*((int *) data), buf);    
 }
 
-static int local_chmod (char *path, int mode)
+static int local_chmod (char *path, unsigned long mode)
 {
     return chmod (path, mode);
 }
@@ -182,7 +178,7 @@ static int local_lseek (void *data, off_t offset, int whence)
     return lseek (fd, offset, whence);
 }
 
-static int local_mknod (char *path, int mode, int dev)
+static int local_mknod (char *path, int mode, dev_t dev)
 {
     return mknod (path, mode, dev);
 }
@@ -192,7 +188,7 @@ static int local_link (char *p1, char *p2)
     return link (p1, p2);
 }
 
-static int local_mkdir (char *path, mode_t mode)
+static int local_mkdir (char *path, unsigned long mode)
 {
     return mkdir (path, mode);
 }
@@ -219,7 +215,7 @@ static void local_free (vfsid id)
 
 static char *local_getlocalcopy (char *path)
 {
-    return strdup (path);
+    return (char *) strdup (path);
 }
 
 static void local_ungetlocalcopy (char *path, char *local, int has_changed)
