@@ -1,12 +1,17 @@
+/*   This has nothing to do with cryptography. */
 #include <config.h>
 #include <stdlib.h>
-#include <util.h>
+#include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
+#ifdef HAVE_PWD
 #include <pwd.h>
+#endif
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#include "util.h"
+#include "diffie/compat.h"
 #include "mad.h"
 
 
@@ -102,7 +107,7 @@ char *canonicalize_pathname (char *path)
 	       ../, which we should leave intact. */
 	    if (path[i + 1] == '.' && (path[i + 2] == PATH_SEP || !path[i + 2])) {
 	        while (--start > -1 && path[start] != PATH_SEP);
-	        if (!strncmp (path + start + 1, "../", 3))
+	        if (!strncmp (path + start + 1, ".." PATH_SEP_STR, 3))
 	            continue;
 	        strcpy (path + start + 1, path + i + 2);
 	        i = start;
@@ -155,13 +160,20 @@ char *copy_strings (const char *first,...)
     return result;
 }
 
+#ifdef HAVE_MAD
+void mad_wipe_password (char *passwd, char *file, int line)
+#else
 void wipe_password (char *passwd)
+#endif
 {
     char *p = passwd;
-    
-    for (;*p ; p++)
-        *p = 0;
+    for (; *p; p++)
+	*p = 0;
+#ifdef HAVE_MAD
+    mad_free (passwd, file, line);
+#else
     free (passwd);
+#endif
 }
 
 char *unix_error_string (int error_num)
@@ -199,7 +211,7 @@ void vfs_default_message_callback (char *p,...)
 
 char *vfs_default_getpasswd_callback (char *a, char *b, char *c, char *host, char *user) /*EK*/
 {
-    return strdup (getpass (b));
+    return (char *) strdup ((char *) getpass (b));
 }
 
 extern char *ftpfs_anonymous_passwd;
