@@ -1,4 +1,5 @@
 /* Virtual File System switch code
+   This has nothing to do with cryptography.
    Copyright (C) 1995 The Free Software Foundation
    
    Written by: 1995 Miguel de Icaza
@@ -21,22 +22,9 @@
 #define _VFS_VFS_C 1
 
 #include <config.h>
-#include <stdio.h>
-#include <stdlib.h>	/* For atol() */
-#include <stdarg.h>
-#include <string.h>
-#include <errno.h>
-#include <sys/types.h>
-#include <malloc.h>
-#include <fcntl.h>
-#include <signal.h>
-#ifdef SCO_FLAVOR
-#include <sys/timeb.h>	/* alex: for struct timeb definition */
-#endif /* SCO_FLAVOR */
-#include <time.h>
-#include <sys/time.h>
-#include "fs.h"
-#include "mad.h"
+#include "net-includes.h"
+#include "diffie/diffie-socket.h"
+#include "diffie/z-socket.h"
 /* #include "dir.h" */
 #include "util.h"
 /* #include "../src/main.h" */
@@ -54,6 +42,8 @@
 #ifdef USE_NETCODE
 #   include "tcputil.h"
 #endif
+#include "diffie/compat.h"
+#include "mad.h"
 
 int cd_symlinks = 1;
 
@@ -78,7 +68,7 @@ static struct {
     vfs  *operations;
 } vfs_file_table [MAX_VFS_FILES];
 
-static int get_bucket ()
+static int get_bucket (void)
 {
     int i;
 
@@ -92,6 +82,36 @@ static int get_bucket ()
     return 0;
 }
 
+#undef min
+#define min(x,y) ((x) <= (y) ? (x) : (y))
+
+long dev_to_long (dev_t dev)
+{
+    long mylong = 0;
+    int endian = 1;
+    unsigned char *e;
+    e = (unsigned char *) &endian;
+    if (*e)
+	memcpy (&mylong, &dev, min (sizeof (mylong), sizeof (dev_t)));
+    else
+	memcpy (&mylong, (char *) &dev + sizeof (dev) - sizeof (mylong), min (sizeof (mylong), sizeof (dev_t)));
+    return mylong;
+}
+
+dev_t long_to_dev (long mylong)
+{
+    dev_t dev;
+    int endian = 1;
+    unsigned char *e;
+    e = (unsigned char *) &endian;
+    memset (&dev, 0, sizeof (dev));
+    if (*e)
+	memcpy (&dev, &mylong, min (sizeof (mylong), sizeof (dev_t)));
+    else
+	memcpy (&dev + sizeof (dev) - sizeof (mylong), &mylong, min (sizeof (mylong), sizeof (dev_t)));
+    return dev;
+}
+
 /* This flag is set on each vfs_type call to say if the path was
    absolute (1) or relative (0) to current dir. */
 
@@ -104,7 +124,7 @@ vfs *vfs_type (char *path)
     vfs_type_absolute = 0;
     vfs = current_vfs;
     
-    if (*path == '/') {
+    if (*path == PATH_SEP) {
     	vfs = &local_vfs_ops;
     	vfs_type_absolute = 1;
     }
@@ -123,7 +143,7 @@ vfs *vfs_type (char *path)
 #endif
     
 #ifdef HAVE_FTPFS
-    if (strncmp (path, "ftp://", 6) == 0){
+    if (strncmp (path, "ftp:", 4) == 0){
         vfs = &ftpfs_vfs_ops;
         vfs_type_absolute = 1;
     }
@@ -160,14 +180,14 @@ char *vfs_path (char *path)
 	return path + 6;
 
 #ifdef USE_NETCODE
-#ifdef USE_NETCODE
+#ifdef HAVE_MCFS
     if (strncmp (path, "mc:", 3) == 0)
-	return strchr (path + 3, '/');
+	return strchr (path + 3, PATH_SEP);
 #endif
 
 #ifdef HAVE_FTPFS
-    if (strncmp (path, "ftp://", 6) == 0)
-	return strchr (path + 6, '/');
+    if (strncmp (path, "ftp:" PATH_SEP_STR PATH_SEP_STR, 6) == 0)
+	return strchr (path + 6, PATH_SEP);
 #endif
 #endif
 
@@ -200,7 +220,7 @@ static struct vfs_stamping *stamps;
  *
  * FIXME: currently this is set to 10 seconds.  We should compute this.
  */
-int vfs_timeouts ()
+int vfs_timeouts (void)
 {
     return stamps ? 10 : 0;
 }
@@ -390,13 +410,14 @@ DIR *mc_opendir (char *dirname)
     char *p = NULL;
     int i = strlen (dirname);
 
-    if (dirname [i - 1] != '/') { 
+    if (dirname [i - 1] != PATH_SEP) { 
     /* We should make possible reading of the root directory in a tar file */
         p = malloc (i + 2);
         strcpy (p, dirname);
-        strcpy (p + i, "/");
+        strcpy (p + i, PATH_SEP_STR);
         dirname = p;
     }
+
     dirname = vfs_canon (dirname);
     vfs = vfs_type (dirname);
 
@@ -526,12 +547,12 @@ char *mc_get_current_wd (char *buffer, int size)
 	mc_stat (buffer, &my_stat);
 	mc_stat (current_dir, &my_stat2);
 	if (my_stat.st_ino != my_stat2.st_ino ||
-	    my_stat.st_dev != my_stat2.st_dev ||
+/* FIXED */    memcmp (&my_stat.st_dev, &my_stat2.st_dev, sizeof (my_stat2.st_dev)) ||
 	    !cd_symlinks) {
 	    if (last_current_dir != current_dir)
 		free (last_current_dir);
 	    free (current_dir);
-	    current_dir = strdup (p);
+	    current_dir = (char *) strdup (p);
 	    last_current_dir = current_dir;
 	    return p;
 	} /* Otherwise we return current_dir below */
@@ -742,7 +763,7 @@ int is_special_prefix (char *path)
         return 1;
 #endif
 #ifdef HAVE_FTPFS
-    if (strncmp (path, "ftp://", 6) == 0)
+    if (strncmp (path, "ftp:" PATH_SEP_STR PATH_SEP_STR, 6) == 0)
         return 1;
 #endif
 #ifdef HAVE_UNDELFS
@@ -752,7 +773,28 @@ int is_special_prefix (char *path)
     return 0;
 }
 
+static char *xvfs_canon (char *path, int *local);
+    
+void char_translate (char *p, char a, char b)
+{
+    for (; *p; p++)
+	if (*p == a)
+	    *p = b;
+}
+
 char *vfs_canon (char *path)
+{
+    int local = 1;
+    path = xvfs_canon (path, &local);
+#if 0
+    if (PATH_SEP == '\\')
+	if (!local)
+	    char_translate (path, '\\', '/');
+#endif
+    return path;
+}
+
+static char *xvfs_canon (char *path, int *local)
 {
     vfs *vfs;
 
@@ -764,10 +806,12 @@ char *vfs_canon (char *path)
 #ifdef USE_NETCODE
 #ifdef HAVE_MCFS
     } else if (strncmp (path, "mc:", 3) == 0) {
+	*local = 0;
 	vfs = &mcfs_vfs_ops;
 #endif
 #ifdef HAVE_FTPFS
-    } else if (strncmp (path, "ftp://", 6) == 0) {
+    } else if (strncmp (path, "ftp:" PATH_SEP_STR PATH_SEP_STR, 6) == 0) {
+	*local = 0;
     	vfs = &ftpfs_vfs_ops;
 #endif
 #endif
@@ -785,8 +829,23 @@ char *vfs_canon (char *path)
     } else if (extfs_prefix_to_type (path) != -1) {
         vfs = &extfs_vfs_ops;
 #endif
-    } else if (*path == '/') { /* Absolute local */
+    } else if (*path == PATH_SEP) {
     	vfs = &local_vfs_ops;
+#ifdef WIN32
+    } else if (strlen (path) >= 2 && path[0] >= 'A' && path[0] <= 'Z' && path[1] == ':') {
+    	vfs = &local_vfs_ops;
+/* HACK: change C:humphrey to C:\humphrey for simplicity */
+	if (path[2] != PATH_SEP) {
+	    char *result;
+	    result = malloc (strlen (path) + 1);
+	    strncpy (result, path, 2);
+	    strcpy (result + 2, PATH_SEP_STR);
+	    strcpy (result + 3, path + 2);
+	    path = vfs_canon (result);
+	    free (result);
+	    return path;
+	}
+#endif
     } else if (*path == '~') { /* Tilde expansion */
     	char *local, *result;
 
@@ -796,30 +855,29 @@ char *vfs_canon (char *path)
 	    free (local);
 	    return result;
 	} else 
-	    return strdup (path);
+	    return (char *) strdup (path);
     } else { /* Relative to current directory */
     	char *local, *result;
 
-	if (current_dir [strlen (current_dir) - 1] == '/')
+	if (current_dir [strlen (current_dir) - 1] == PATH_SEP)
 	    local = copy_strings (current_dir, path, NULL);
 	else
-	    local = copy_strings (current_dir, "/", path, NULL);
-
+	    local = copy_strings (current_dir, PATH_SEP_STR, path, NULL);
 	result = vfs_canon (local);
 	free (local);
 	return result;
     }
-    
+
     if (vfs == &local_vfs_ops)
 /* NOTE: No local:/ is necessary at the moment, we are handling absolute paths
          at the moment, even it is impossible, since it doesn't work then at all. */    
-	return strdup (canonicalize_pathname (path));
+	return (char *) strdup (canonicalize_pathname (path));
 #ifdef USE_EXT2FSLIB
     if (vfs == &undelfs_vfs_ops) {
 	int i = strlen (path);
 	if ((path[i - 1] == '.' && path[i - 2] == '.') ||
-	(path[i - 1] == '/' && path[i - 2] == '.' && path[i - 3] == '.'))
-	    return strdup ("/"); /* we want to go out of the undelfs ->
+	(path[i - 1] == PATH_SEP && path[i - 2] == '.' && path[i - 3] == '.'))
+	    return strdup (PATH_SEP_STR); /* we want to go out of the undelfs ->
 	                            go to the root directory */
 	else
 	    return strdup (path);
@@ -842,15 +900,15 @@ char *vfs_canon (char *path)
 	    prefixsh = 6;
 	    isftp = 1;
 	} else {
-	    if (path[3] == '/' && path[4] == '/')
+	    if (path[3] == PATH_SEP && path[4] == PATH_SEP)
 		prefixsh = 5;
 	    else
 		prefixsh = 3;
 	    isftp = 0;
 	}
 	
-	q = strdup (path);
-	p = strchr (q + prefixsh, '/');
+	q = (char *) strdup (path);
+	p = strchr (q + prefixsh, PATH_SEP);
 	
 	if (p != NULL) {
 	    if (p [0] && p [1] == '~') { /* Tilde expansion */
@@ -858,16 +916,16 @@ char *vfs_canon (char *path)
 	    	if (isftp) {
 	            s = ftpfs_gethome (q);
 	            if (s == NULL)
-	                s = "/";
+	                s = PATH_SEP_STR;
 	        }
 #ifdef HAVE_MCFS
 		else {
 	            s = mcfs_gethome (q);
 	            if (s == NULL)
-	                s = strdup ("/");
+	                s = (char *) strdup (PATH_SEP_STR);
 	        }
 #endif
-	        if (p [2] == '/')
+	        if (p [2] == PATH_SEP)
 	            p += 3;
 	        else
 	            p += 2;
@@ -876,7 +934,7 @@ char *vfs_canon (char *path)
 	            free (s);
 	        free (q);
 	        q = r;
-	        p = strchr (q + prefixsh, '/');
+	        p = strchr (q + prefixsh, PATH_SEP);
 	    }
 	    p++;
 	    if (p [0])
@@ -884,7 +942,7 @@ char *vfs_canon (char *path)
 	    else
 		p = ".";
 	    
-	    if (!strncmp (p, "..", 2) && (p [2] == '/' || !p [2])) {
+	    if (!strncmp (p, "..", 2) && (p [2] == PATH_SEP || !p [2])) {
 	    	*p = 0;
 #ifdef HAVE_MCFS
 	        r = isftp ? ftpfs_getupdir (q) : mcfs_getupdir (q);
@@ -892,9 +950,9 @@ char *vfs_canon (char *path)
 	        r = ftpfs_getupdir (q);
 #endif
 	        if (r == NULL)
-	            r = "/";
+	            r = PATH_SEP_STR;
 	        p += 2;
-	        if (*p == '/')
+	        if (*p == PATH_SEP)
 	            p++;
 	        r = copy_strings (r, p, NULL);
 	        free (q);
@@ -931,18 +989,18 @@ char *vfs_canon (char *path)
 #endif
         }
         if (local_path == NULL) {
-            return strdup ("/"); /* For error cases cd to the root dir */
+            return strdup (PATH_SEP_STR); /* For error cases cd to the root dir */
         }
-        p = (*local_path == '/') ? local_path + 1 : local_path;
+        p = (*local_path == PATH_SEP) ? local_path + 1 : local_path;
         if (*p)
             canonicalize_pathname (p);
         if (!strcmp (p, ".")) /* This is an incorrect result of canonicalization,
                                  if we have p = somedir/.. */
 	    *p = 0;                                 
-        if (p [0] != '.' || p [1] != '.' || (p [2] && p [2] != '/')) { 
+        if (p [0] != '.' || p [1] != '.' || (p [2] && p [2] != PATH_SEP)) { 
             /* I.e. p doesn't start with the .. directory */
             if (istar)
-	        result = copy_strings ("tar:", arc_name, "/", p, 0);
+	        result = copy_strings ("tar:", arc_name, PATH_SEP_STR, p, 0);
 	    else {
 		/* Two chances: the extfs requires an archive to
 		 * use or it does not need it:
@@ -951,7 +1009,7 @@ char *vfs_canon (char *path)
 		/* If it does need an archive... */
 		if (arc_name && *arc_name)
 		    result = copy_strings
-			(extfs_prefix, ":", arc_name, "/", p, 0);
+			(extfs_prefix, ":", arc_name, PATH_SEP_STR, p, 0);
 		else 
 		    result = canonicalize_pathname (strdup(path));
 	    }
@@ -964,15 +1022,15 @@ char *vfs_canon (char *path)
 	    if (!istar && (!arc_name || !*arc_name)) {
 	        free (local_path);
 	        if (arc_name) free (arc_name);
-	        return strdup ("/");
+	        return strdup (PATH_SEP_STR);
 	    }
-	    q = strrchr (arc_name, '/');
+	    q = strrchr (arc_name, PATH_SEP);
 	    if (q == NULL)
 	    	q = arc_name; /* Should not happen */
 	    else
 	    	q++; 
 	    *q = 0;
-	    if (p [2] == '/')
+	    if (p [2] == PATH_SEP)
 	    	p += 3;
 	    else
 	    	p += 2;
@@ -995,8 +1053,8 @@ vfsid vfs_ncs_getid (vfs *nvfs, char *dir, struct vfs_stamping **par)
     vfsid nvfsid;
     int freeit = 0;
 
-    if (dir [strlen (dir) - 1] != '/') {
-        dir = copy_strings (dir, "/", NULL);    
+    if (dir [strlen (dir) - 1] != PATH_SEP) {
+        dir = copy_strings (dir, PATH_SEP_STR, NULL);    
         freeit = 1;
     }
     nvfsid = (*nvfs->getid)(dir, par);
@@ -1135,16 +1193,16 @@ int mc_chdir (char *path)
     vfsid oldvfsid;
     struct vfs_stamping *parent;
 
-    if (path [i - 1] != '/') {
+    if (path [i - 1] != PATH_SEP) {
     /* We should make possible reading of the root directory in a tar archive */
         p = malloc (i + 2);
         strcpy (p, path);
-        strcpy (p + i, "/");
+        strcpy (p + i, PATH_SEP_STR);
         path = p;
     }
     
     if (!current_dir) {
-	current_dir = strdup ("/"); /* Note: we should set current_dir
+	current_dir = (char *) strdup (PATH_SEP_STR); /* Note: we should set current_dir
 	                               on startup to a reasonable pwd.
 	                               It cannot be relative, i.e. no
 	                               . or anything else except
@@ -1180,7 +1238,7 @@ int mc_chdir (char *path)
     }
     if (*current_dir) {
 	p = strchr (current_dir, 0) - 1;
-	if (*p == '/' && p > current_dir)
+	if (*p == PATH_SEP && p > current_dir)
 	    *p = 0; /* Sometimes we assume no trailing slash on cwd */
     }
     return result;
@@ -1249,7 +1307,7 @@ static void vfs_setup_wd (void)
     return;
 }
 
-int mc_mkdir (char *path, mode_t mode)
+int mc_mkdir (char *path, unsigned long mode)
 {
     vfs *vfs;
     int result;
@@ -1281,7 +1339,7 @@ int mc_rmdir (char *path)
     return result;
 }
 
-int mc_mknod (char *path, int mode, int dev)
+int mc_mknod (char *path, int mode, dev_t dev)
 {
     vfs *vfs;
     int result;
@@ -1360,12 +1418,12 @@ char *mc_def_getlocalcopy (char *filename)
     if (fdin == -1)
         return NULL;
     tmp = tmpnam(NULL);
-    fdout = creat (tmp, 0600);
+    fdout = creat (tmp, S_IWUSR | S_IRUSR);
     if (fdout == -1) {
         mc_close (fdin);
         return NULL;
     }
-    tmp = strdup (tmp);
+    tmp = (char *) strdup (tmp);
     while ((i = mc_read (fdin, buffer, sizeof (buffer))) == sizeof (buffer)) {
         write (fdout, buffer, i);
     }
@@ -1442,15 +1500,15 @@ inline int timeoutcmp (struct timeval *t1, struct timeval *t2)
 	    || ((t1->tv_sec == t2->tv_sec) && (t1->tv_usec <= t2->tv_usec)));
 }
 
-void vfs_timeout_handler ()
+void vfs_timeout_handler (void)
 {
-    struct timeval time;
+    struct timeval t;
     struct vfs_stamping *stamp, *st;
 
-    gettimeofday (&time, NULL);
-    time.tv_sec -= vfs_timeout;
+    gettimeofday (&t, NULL);
+    t.tv_sec -= vfs_timeout;
     for (stamp = stamps; stamp != NULL;) {
-        if (timeoutcmp (&stamp->time, &time)) {
+        if (timeoutcmp (&stamp->time, &t)) {
             st = stamp->next;
             (*stamp->v->free) (stamp->id);
 	    vfs_rmstamp (stamp->v, stamp->id, 0);
@@ -1464,12 +1522,27 @@ void vfs_init (void)
 {
     time_t current_time;
     struct tm *t;
-    
+
+#ifdef WIN32
+    WSADATA wsa_data;
+/* FIXME: should return proper error */
+    if (WSAStartup(0x0101, &wsa_data))
+	message_1s (1, " Error ", " Could not initialise windows socket library ");
+    if (wsa_data.wVersion != 0x0101)
+	message_1s (1, " Error ", " Could not initialise windows socket library to version 1.1 ");
+#endif
+
     current_time = time (NULL);
     t = localtime (&current_time);
     current_mon = t->tm_mon;
     current_year = t->tm_year;
-    
+
+
+
+#ifdef HAVE_MCFS
+    mcfs_init ();
+#endif
+
 #ifdef USE_NETCODE
     tcp_init();
 #ifdef HAVE_FTPFS
@@ -1538,6 +1611,9 @@ void vfs_shut (void)
 #endif
 #endif
 
+#ifdef WIN32
+    WSACleanup ();
+#endif
 }
 
 /* These ones grab information from the VFS
@@ -1628,65 +1704,65 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
 	    return 0;
 	/* Should parse here the Notwell permissions :) */
 	if (S_ISDIR (s->st_mode))
-	    s->st_mode |= (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IXUSR | S_IXGRP | S_IXOTH);
+	    s->st_mode |= double_mode (0755);
 	else
-	    s->st_mode |= (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR);
+	    s->st_mode |= double_mode (0644);
 	p += 9;
     } else {
 	switch (*(p++)) {
-	 case 'r': s->st_mode |= 0400; break;
+	 case 'r': s->st_mode |= double_mode (0400); break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'w': s->st_mode |= 0200; break;
+	 case 'w': s->st_mode |= double_mode (0200); break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'x': s->st_mode |= 0100; break;
-	 case 's': s->st_mode |= 0100 | S_ISUID; break;
+	 case 'x': s->st_mode |= double_mode (0100); break;
+	 case 's': s->st_mode |= double_mode (0100) | S_ISUID; break;
 	 case 'S': s->st_mode |= S_ISUID; break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'r': s->st_mode |= 0040; break;
+	 case 'r': s->st_mode |= double_mode (0040); break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'w': s->st_mode |= 0020; break;
+	 case 'w': s->st_mode |= double_mode (0020); break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'x': s->st_mode |= 0010; break;
-	 case 's': s->st_mode |= 0010 | S_ISGID; break;
+	 case 'x': s->st_mode |= double_mode (0010); break;
+	 case 's': s->st_mode |= double_mode (0010) | S_ISGID; break;
 	 case 'S': s->st_mode |= S_ISGID; break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'r': s->st_mode |= 0004; break;
+	 case 'r': s->st_mode |= double_mode (0004); break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'w': s->st_mode |= 0002; break;
+	 case 'w': s->st_mode |= double_mode (0002); break;
 	 case '-': break;
 	 default: return 0;
 	}
 	switch (*(p++)) {
-	 case 'x': s->st_mode |= 0001; break;
-	 case 't': s->st_mode |= 0001 | S_ISVTX; break;
+	 case 'x': s->st_mode |= double_mode (0001); break;
+	 case 't': s->st_mode |= double_mode (0001) | S_ISVTX; break;
 	 case 'T': s->st_mode |= S_ISVTX; break;
 	 case '-': break;
 	 default: return 0;
 	}
     }
 	
-    p_copy = strdup (p);
+    p_copy = (char *) strdup (p);
     num_cols = split_text (p);
 
     s->st_nlink = atol (columns [0]);
@@ -1707,8 +1783,8 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
     if (idx == 6 || (idx == 5 && !S_ISCHR (s->st_mode) && !S_ISBLK (s->st_mode)))
         free_and_return (0);
     if (idx < 5) {
-        char *p = strchr(columns [idx - 1], ',');
-        if (p && p[1] >= '0' && p[1] <= '9')
+        char *xp = strchr(columns [idx - 1], ',');
+        if (xp && xp[1] >= '0' && xp[1] <= '9')
             isconc = 1;
     }
     if (idx == 3 || (idx == 4 && !isconc && (S_ISCHR(s->st_mode) || S_ISBLK (s->st_mode))))
@@ -1722,34 +1798,36 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
     }
 
     if (S_ISCHR (s->st_mode) || S_ISBLK (s->st_mode)) {
-        char *p;
+        char *xp;
+	long strdev;
 	if (!is_num (idx2))
 	    free_and_return (0);
 #ifdef HAVE_ST_RDEV
-	s->st_rdev = (atol (columns [idx2]) & 0xff) << 8;
+	strdev = (atol (columns [idx2]) & 0xff) << 8;
 #endif
 	if (isconc) {
-	    p = strchr (columns [idx2], ',');
-	    if (!p || p [1] < '0' || p [1] > '9')
+	    xp = strchr (columns [idx2], ',');
+	    if (!xp || xp [1] < '0' || xp [1] > '9')
 	        free_and_return (0);
-	    p++;
+	    xp++;
 	} else {
-	    p = columns [idx2 + 1];
+	    xp = columns [idx2 + 1];
 	    if (!is_num (idx2+1))
 	        free_and_return (0);
 	}
 	
 #ifdef HAVE_ST_RDEV
-	s->st_rdev |= (atol (p) & 0xff);
+	strdev |= (atol (xp) & 0xff);
+	s->st_rdev = long_to_dev (strdev);
 #endif
 	s->st_size = 0;
     } else {
 	if (!is_num (idx2))
 	    free_and_return (0);
-	
+/* FIXED: check all st_dev and st_rdev especially on mcfs rpc calls */	
 	s->st_size = (size_t) atol (columns [idx2]);
 #ifdef HAVE_ST_RDEV
-	s->st_rdev = 0;
+	s->st_rdev = long_to_dev (0);
 #endif
     }
     
@@ -1848,7 +1926,7 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
         s->st_mtime = 0;
     s->st_atime = s->st_mtime;
     s->st_ctime = s->st_mtime;
-    s->st_dev = 0;
+    s->st_dev = long_to_dev (0);
     s->st_ino = 0;
 #ifdef HAVE_ST_BLKSIZE
     s->st_blksize = 512;
@@ -1866,25 +1944,25 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
     if (((S_ISLNK (s->st_mode) || 
         (num_cols == idx + 3 && s->st_nlink > 1))) /* Maybe a hardlink? (in extfs) */
         && idx2){
-	int p;
-	char *s;
+	int xp;
+	char *xs;
 	    
 	if (filename) {
-	    p = column_ptr [idx2] - column_ptr [idx];
-	    s = malloc (p);
-	    strncpy (s, p_copy + column_ptr [idx], p - 1);
-	    s[p - 1] = '\0';
-	    *filename = s;
+	    xp = column_ptr [idx2] - column_ptr [idx];
+	    xs = malloc (xp);
+	    strncpy (xs, p_copy + column_ptr [idx], xp - 1);
+	    xs[xp - 1] = '\0';
+	    *filename = xs;
 	}
 	if (linkname) {
-	    s = strdup (p_copy + column_ptr [idx2+1]);
-	    p = strlen (s);
-	    if (s [p-1] == '\r' || s [p-1] == '\n')
-		s [p-1] = 0;
-	    if (s [p-2] == '\r' || s [p-2] == '\n')
-		s [p-2] = 0;
+	    xs = (char *) strdup (p_copy + column_ptr [idx2+1]);
+	    xp = strlen (xs);
+	    if (xs [xp-1] == '\r' || xs[xp-1] == '\n')
+		xs [xp-1] = 0;
+	    if (xs [xp-2] == '\r' || xs[xp-2] == '\n')
+		xs [xp-2] = 0;
 		
-	    *linkname = s;
+	    *linkname = xs;
 	}
     } else {
 	/* Extract the filename from the string copy, not from the columns
@@ -1894,17 +1972,17 @@ int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
 	    /* 
 	    *filename = strdup (columns [idx++]);
 	    */
-	    int p;
-	    char *s;
+	    int xp;
+	    char *xs;
 	    
-	    s = strdup (p_copy + column_ptr [idx++]);
-	    p = strlen (s);
-	    if (s [p-1] == '\r' || s [p-1] == '\n')
-		s [p-1] = 0;
-	    if (s [p-2] == '\r' || s [p-2] == '\n')
-		s [p-2] = 0;
+	    xs = (char *) strdup (p_copy + column_ptr [idx++]);
+	    xp = strlen (xs);
+	    if (xs [xp-1] == '\r' || xs[xp-1] == '\n')
+		xs [xp-1] = 0;
+	    if (xs [xp-2] == '\r' || xs[xp-2] == '\n')
+		xs [xp-2] = 0;
 	    
-	    *filename = s;
+	    *filename = xs;
 	}
 	if (linkname)
 	    *linkname = NULL;
